;----------------------------------------------------------------------------
; 1536-color example program
;
; Written by Charles MacDonald
; E-mail: cgfm2@hotmail.com
; WWW: http://cgfm2.emuviews.com
;
; Works on the real thing. But if you're using an emulator make sure it
; supports displays with more than 256 colors.
;
; Using raster effects, 512 colors are shown in four 16x128 columns.
; The columns are duplicated three times, with each group being
; displayed in half intensity, normal intensity, and double intensity.
; The end result is a total of 1536 colors displayed on-screen at once.
;----------------------------------------------------------------------------

; Set to use 40-cell mode
H40                     equ     1

; RAM location that the VBlank handler sets
VBL_FLAG                equ     $FF0004

; Load the VDP DMA registers with a 23-bit address
setdmaaddr              MACRO
                        move.w  #$9500|((\1>>1)&$FF), (a1)
                        move.w  #$9600|((\1>>9)&$FF), (a1)
                        move.w  #$9700|((\1>>17)&$7F), (a1)
                        ENDM

; Our standard include files
                        include init.inc
                        include vdp.inc
_start:
                        ; Set up the VDP
                        jsr     _vdp_init

                        ; Make tile graphics
                        move.l  #$40000000, (a1)
                        move.l  #$0F, d0
                        moveq   #0, d2
        makechar:       move.l  #$07, d1
        makeline:       move.l  d2, (a0)
                        dbra    d1, makeline
                        addi.l  #$11111111, d2
                        dbra    d0, makechar

                        ; Make hilight sprite patterns
                        ; Using #$E0E0E0E0 or #$EFEFEFEF looks good too,
                        ; but the effect is lost in 32 cell mode.
                        move.l  #$7F, d0
    make_s:             move.l  #$EEEEEEEE, (a0)
                        dbra    d0, make_s

                        ; Make background tile
                        move.l  #$40000000, (a1)
                        move.l  #$01100110, (a0)
                        move.l  #$10000001, (a0)
                        move.l  #$10011001, (a0)
                        move.l  #$00100100, (a0)
                        move.l  #$00100100, (a0)
                        move.l  #$10011001, (a0)
                        move.l  #$10000001, (a0)
                        move.l  #$01100110, (a0)

                        ; Make test pattern
                        move.l  #$40000001, (a1)
                        move.l  #$0F, d0                ; 16 rows
        make_row:       move.l  #$00, d1                ; 12 cols
                        move.w  #$0004, d2
        make_col:       move.w  d2, (a0)
                        move.w  d2, (a0)
                        addi.w  #$0001, d2
                        andi.w  #$8003, d2
                        ori.w   #$0004, d2
                        cmpi.l  #$07, d1
                        bne.s   no_pri
                        ori.w   #$8000, d2
        no_pri:         addi.l  #1, d1
                        cmpi.l  #12, d1
                        bne.s   make_col
                        move.l  #39, d1
        fill_col:       move.w  #$0000, (a0)
                        dbra    d1, fill_col
                        dbra    d0, make_row

                        ; Center plane A 
                        move.l  #$40000010, (a1)
                        move.w  #$FFD0, (a0)
                        move.l  #$70000001, (a1)
                        move.w  #$0040, (a0)

                        ; Draw sprites
                        move.l  #$78000001, (a1)
                        move.w  #$0F01, d1
                        move.w  #$E010, d2
                        move.w  #$0100, d3
                        move.l  #$03, d5
    obj2l:              move.l  #$0003, d4
                        move.w  #$00B0, d0
    obj1l:              move.w  d0, (a0)
                        move.w  d1, (a0)
                        move.w  d2, (a0)
                        move.w  d3, (a0)
                        addi.w  #$0020, d0
                        addi.b  #$01, d1
                        dbra    d4, obj1l
                        addi.w  #$0020, d3
                        dbra    d5, obj2l

                        ; Set up registers for following frames
                        move.w  #$8114, d1       ; Screen off/DMA on value
                        move.w  #$8164, d2       ; Screen on/DMA off value
                        move.w  #$8C89, (a1)     ; 320x224, S/TE on
                        move.w  #$9400, (a1)     ; Clear MSB of DMA length
                        lea     $FF0000, a2      ; Location of VDP command
                        move.l  #$C0080080, (a2) ; Our VDP command

                        ; Now we want to enable frame interrupts for the VDP
                        ; and level 6 interrupts for the 68000. However,
                        ; at any time the VDP reaches line $E0 in mode 5
                        ; and it can't generate an interrupt, it marks
                        ; the interrupt as pending. As soon as interrupts
                        ; are enabled, the missed interrupt will be
                        ; generated. You can check for the occurance of
                        ; a pending interrupt by testing bit 7 of the
                        ; status flags.
                        ;
                        ; The bad part is that you can't clear a pending
                        ; interrupt. The only solution is to let the
                        ; interrupt execute. The SMS could clear pending
                        ; interrupts by reading the control port, so
                        ; we instead switch into mode 4 (SMS display mode),
                        ; read the control port, and then switch back.
                        ;
                        ; This trick may not work on later versions of
                        ; the Genesis which don't support mode 4. In that
                        ; case, you'd want to let the interrupt routine
                        ; execute, and then reset any variables or things
                        ; it modified before entering the main loop.
                        ; Not exactly a good solution.
                        ;
                        ; The only reason we know that an interrupt is
                        ; pending is that by the time the 68000 reaches
                        ; this code, we've done a lot of setup, and have
                        ; wasted a few frames so far. I include a check
                        ; to see if there are any pending interrupts just
                        ; to be proper about it, but it's not necessary.

                        btst    #7, $C00005     ; Any pending interrupts?
                        beq.s   no_pending      ; Nope..
                        move.w  #$8100, (a1)    ; Select mode 4
                        tst.w   (a1)            ; Clear the pending interrupt
    no_pending:         move.w  #$8164, (a1)    ; Mode 5, frame ints. on
                        move.w  #$2500, sr      ; Enable level 6 interrupt
                        clr.b   VBL_FLAG        ; Clear our VBlank flag

;----------------------------------------------------------------------------

                        ; Wait until we're at the end of the frame.
                        ; More specifically, wait until the VBlank handler
                        ; has been called at line $E0.
                        ; (The previous version waited for line $E0, but
                        ; this could be missed if the VBlank handler took
                        ; longer than one line to execute.. which it does!)
    frame_top:
                        cmpi.b  #$FF, $FF0004  ; Has our VBlank flag changed?
                        bne.s   frame_top
                        clr.b   VBL_FLAG       ; We're done with it

                        ; To shorten the amount of processing that has to
                        ; be done on each line, we set up the DMA address
                        ; here, and then allow the DMA source address
                        ; registers to increment as we transfer four words
                        ; of ROM data to CRAM on each of 128 scanlines.
                        ; This saves us from having to re-enter the source
                        ; address on every line.
                        ;
                        ; After writing the last byte of the source address
                        ; to register #97, the VDP will always expect the
                        ; next thing written to be the command word.
                        ; To satisfy this behavior we have to do a DMA
                        ; transfer, so we use this to set up the background
                        ; color. It's more like a 'dummy' operation, which is
                        ; just to keep the VDP happy. If we just wrote to
                        ; register #97, without doing DMA, our write on the
                        ; next line to turn the display off would violate
                        ; this rule. (since it's not the command word to
                        ; start DMA, but rather a register write)
                              
                        move.l  #$81549302, (a1)    ; Display off, DMA on,
                                                    ; source count = 2 words
                        setdmaaddr paldata          ; Set source address
                                                    ; of palette data
                        move.l  #$C0000080, -(a7)   ; Write the command word
                        move.l  (a7)+, (a1)         ; CPU is halted here...
                        move.l  #$81449304, (a1)    ; Display on, DMA off,
                                                    ; and set the source
                                                    ; count for the next line
                                                    ; in advance.

                        ; Wait for line 48, where the color changes begin

    wstart:             cmpi.b  #48, $C00008
                        bne.s   wstart
    wline:
                        ; The VDP increments the V counter at the end of
                        ; the current scanline, before horizontal blanking.
                        ; That means we aren't really at the next scanline,
                        ; (despite what the V counter says), but instead
                        ; are at the small bit of space between this line
                        ; and the next one. This is where we update the
                        ; palette, to reduce on-screen flicker.
                        ;
                        ; When the palette is changed, the VDP (for
                        ; whatever reason) will show one pixel with the
                        ; new color value written, over any underlying
                        ; display data. This is what those of you with VGA
                        ; programming experience would call snow or flicker.
                        ; Updating the palette outside of the active display
                        ; period helps reduce this. You can see this effect
                        ; in a game that changes the palette mid-frame,
                        ; good examples include Samurai Shodown and any of
                        ; the Sonic games.
                        ;
                        ; What's really weird is that regardless of where
                        ; the palette is set during this period, the VDP
                        ; will still show some flicker anyway. I found
                        ; that turning the display off (though we aren't
                        ; even in the active display period!) stops this.
                        ;
                        ; Of course, this means that you have to turn
                        ; the screen off, do DMA, and turn it back on all
                        ; before the next line starts. As it turns out,
                        ; we have *just* enough time to pull that off.
                        ;
                        ; If you need to buy some extra time, you can
                        ; set bit 5 of VDP register #0, which blanks
                        ; the leftmost 8 pixels of the display, and hides
                        ; flicker. But I wanted to write something that
                        ; was fast enough to not need that. :)

                        move.w  d1, (a1)        ; Display off, DMA on
                        move.l  (a2), (a1)      ; Write the command word
                                                ; to start DMA.                                              
                        move.w  d2, (a1)        ; Turn the screen back on.
                        move.w  #$9304, (a1)    ; Set the source count
                                                ; for the next line

                        ; Wait for the next line.

                        move.b  $C00008, d0
    wnext:              cmp.b   $C00008, d0
                        beq.s   wnext

                        ; If we're at the end of the block of graphics we
                        ; want to change the colors of, then quit

                        cmpi.b  #176, d0
                        bne.s   wline

                        jmp     frame_top

;----------------------------------------------------------------------------
_frame_int:
                        move.l  d0-d7, -(a7)

                        ; Set the VBlank flag
                        st.b    VBL_FLAG

                        ; Animate the background
                        move.l  #$00000000, (a1)
                        rept    4
                        nop
                        endr
                        move.l  (a0), d0
                        move.l  (a0), d1
                        move.l  (a0), d2
                        move.l  (a0), d3
                        move.l  (a0), d4
                        move.l  (a0), d5
                        move.l  (a0), d6
                        move.l  (a0), d7
                        rol.l   #4, d0
                        rol.l   #4, d1
                        rol.l   #4, d2
                        rol.l   #4, d3
                        rol.l   #4, d4
                        rol.l   #4, d5
                        rol.l   #4, d6
                        rol.l   #4, d7
                        move.l  #$40000000, (a1)
                        move.l  d7, (a0)
                        move.l  d0, (a0)
                        move.l  d1, (a0)
                        move.l  d2, (a0)
                        move.l  d3, (a0)
                        move.l  d4, (a0)
                        move.l  d5, (a0)
                        move.l  d6, (a0)

                        move.l  (a7)+, d0-d7
                        rte

;----------------------------------------------------------------------------
_line_int:        
                        rte

;----------------------------------------------------------------------------

paldata:                               ; block of 512 palette values

    dc.w    $0080, $0060               ; dummy data for background
	dc.w	$0000, $0400, $0800, $0C00 
	dc.w	$0002, $0402, $0802, $0C02 
	dc.w	$0004, $0404, $0804, $0C04 
	dc.w	$0006, $0406, $0806, $0C06 
	dc.w	$0008, $0408, $0808, $0C08 
	dc.w	$000A, $040A, $080A, $0C0A 
	dc.w	$000C, $040C, $080C, $0C0C 
	dc.w	$000E, $040E, $080E, $0C0E 
	dc.w	$0020, $0420, $0820, $0C20 
	dc.w	$0022, $0422, $0822, $0C22 
	dc.w	$0024, $0424, $0824, $0C24 
	dc.w	$0026, $0426, $0826, $0C26 
	dc.w	$0028, $0428, $0828, $0C28 
	dc.w	$002A, $042A, $082A, $0C2A 
	dc.w	$002C, $042C, $082C, $0C2C 
	dc.w	$002E, $042E, $082E, $0C2E 
	dc.w	$0040, $0440, $0840, $0C40 
	dc.w	$0042, $0442, $0842, $0C42 
	dc.w	$0044, $0444, $0844, $0C44 
	dc.w	$0046, $0446, $0846, $0C46 
	dc.w	$0048, $0448, $0848, $0C48 
	dc.w	$004A, $044A, $084A, $0C4A 
	dc.w	$004C, $044C, $084C, $0C4C 
	dc.w	$004E, $044E, $084E, $0C4E 
	dc.w	$0060, $0460, $0860, $0C60 
	dc.w	$0062, $0462, $0862, $0C62 
	dc.w	$0064, $0464, $0864, $0C64 
	dc.w	$0066, $0466, $0866, $0C66 
	dc.w	$0068, $0468, $0868, $0C68 
	dc.w	$006A, $046A, $086A, $0C6A 
	dc.w	$006C, $046C, $086C, $0C6C 
	dc.w	$006E, $046E, $086E, $0C6E 
	dc.w	$0080, $0480, $0880, $0C80 
	dc.w	$0082, $0482, $0882, $0C82 
	dc.w	$0084, $0484, $0884, $0C84 
	dc.w	$0086, $0486, $0886, $0C86 
	dc.w	$0088, $0488, $0888, $0C88 
	dc.w	$008A, $048A, $088A, $0C8A 
	dc.w	$008C, $048C, $088C, $0C8C 
	dc.w	$008E, $048E, $088E, $0C8E 
	dc.w	$00A0, $04A0, $08A0, $0CA0 
	dc.w	$00A2, $04A2, $08A2, $0CA2 
	dc.w	$00A4, $04A4, $08A4, $0CA4 
	dc.w	$00A6, $04A6, $08A6, $0CA6 
	dc.w	$00A8, $04A8, $08A8, $0CA8 
	dc.w	$00AA, $04AA, $08AA, $0CAA 
	dc.w	$00AC, $04AC, $08AC, $0CAC 
	dc.w	$00AE, $04AE, $08AE, $0CAE 
	dc.w	$00C0, $04C0, $08C0, $0CC0 
	dc.w	$00C2, $04C2, $08C2, $0CC2 
	dc.w	$00C4, $04C4, $08C4, $0CC4 
	dc.w	$00C6, $04C6, $08C6, $0CC6 
	dc.w	$00C8, $04C8, $08C8, $0CC8 
	dc.w	$00CA, $04CA, $08CA, $0CCA 
	dc.w	$00CC, $04CC, $08CC, $0CCC 
	dc.w	$00CE, $04CE, $08CE, $0CCE 
	dc.w	$00E0, $04E0, $08E0, $0CE0 
	dc.w	$00E2, $04E2, $08E2, $0CE2 
	dc.w	$00E4, $04E4, $08E4, $0CE4 
	dc.w	$00E6, $04E6, $08E6, $0CE6 
	dc.w	$00E8, $04E8, $08E8, $0CE8 
	dc.w	$00EA, $04EA, $08EA, $0CEA 
	dc.w	$00EC, $04EC, $08EC, $0CEC 
	dc.w	$00EE, $04EE, $08EE, $0CEE 
	dc.w	$0200, $0600, $0A00, $0E00 
	dc.w	$0202, $0602, $0A02, $0E02 
	dc.w	$0204, $0604, $0A04, $0E04 
	dc.w	$0206, $0606, $0A06, $0E06 
	dc.w	$0208, $0608, $0A08, $0E08 
	dc.w	$020A, $060A, $0A0A, $0E0A 
	dc.w	$020C, $060C, $0A0C, $0E0C 
	dc.w	$020E, $060E, $0A0E, $0E0E 
	dc.w	$0220, $0620, $0A20, $0E20 
	dc.w	$0222, $0622, $0A22, $0E22 
	dc.w	$0224, $0624, $0A24, $0E24 
	dc.w	$0226, $0626, $0A26, $0E26 
	dc.w	$0228, $0628, $0A28, $0E28 
	dc.w	$022A, $062A, $0A2A, $0E2A 
	dc.w	$022C, $062C, $0A2C, $0E2C 
	dc.w	$022E, $062E, $0A2E, $0E2E 
	dc.w	$0240, $0640, $0A40, $0E40 
	dc.w	$0242, $0642, $0A42, $0E42 
	dc.w	$0244, $0644, $0A44, $0E44 
	dc.w	$0246, $0646, $0A46, $0E46 
	dc.w	$0248, $0648, $0A48, $0E48 
	dc.w	$024A, $064A, $0A4A, $0E4A 
	dc.w	$024C, $064C, $0A4C, $0E4C 
	dc.w	$024E, $064E, $0A4E, $0E4E 
	dc.w	$0260, $0660, $0A60, $0E60 
	dc.w	$0262, $0662, $0A62, $0E62 
	dc.w	$0264, $0664, $0A64, $0E64 
	dc.w	$0266, $0666, $0A66, $0E66 
	dc.w	$0268, $0668, $0A68, $0E68 
	dc.w	$026A, $066A, $0A6A, $0E6A 
	dc.w	$026C, $066C, $0A6C, $0E6C 
	dc.w	$026E, $066E, $0A6E, $0E6E 
	dc.w	$0280, $0680, $0A80, $0E80 
	dc.w	$0282, $0682, $0A82, $0E82 
	dc.w	$0284, $0684, $0A84, $0E84 
	dc.w	$0286, $0686, $0A86, $0E86 
	dc.w	$0288, $0688, $0A88, $0E88 
	dc.w	$028A, $068A, $0A8A, $0E8A 
	dc.w	$028C, $068C, $0A8C, $0E8C 
	dc.w	$028E, $068E, $0A8E, $0E8E 
	dc.w	$02A0, $06A0, $0AA0, $0EA0 
	dc.w	$02A2, $06A2, $0AA2, $0EA2 
	dc.w	$02A4, $06A4, $0AA4, $0EA4 
	dc.w	$02A6, $06A6, $0AA6, $0EA6 
	dc.w	$02A8, $06A8, $0AA8, $0EA8 
	dc.w	$02AA, $06AA, $0AAA, $0EAA 
	dc.w	$02AC, $06AC, $0AAC, $0EAC 
	dc.w	$02AE, $06AE, $0AAE, $0EAE 
	dc.w	$02C0, $06C0, $0AC0, $0EC0 
	dc.w	$02C2, $06C2, $0AC2, $0EC2 
	dc.w	$02C4, $06C4, $0AC4, $0EC4 
	dc.w	$02C6, $06C6, $0AC6, $0EC6 
	dc.w	$02C8, $06C8, $0AC8, $0EC8 
	dc.w	$02CA, $06CA, $0ACA, $0ECA 
	dc.w	$02CC, $06CC, $0ACC, $0ECC 
	dc.w	$02CE, $06CE, $0ACE, $0ECE 
	dc.w	$02E0, $06E0, $0AE0, $0EE0 
	dc.w	$02E2, $06E2, $0AE2, $0EE2 
	dc.w	$02E4, $06E4, $0AE4, $0EE4 
	dc.w	$02E6, $06E6, $0AE6, $0EE6 
	dc.w	$02E8, $06E8, $0AE8, $0EE8 
	dc.w	$02EA, $06EA, $0AEA, $0EEA 
	dc.w	$02EC, $06EC, $0AEC, $0EEC 
	dc.w	$02EE, $06EE, $0AEE, $0EEE

                        cnop    0, $4000

